//
// Created by philipp on 02.01.18.
//

#ifndef NETWORKINGLIB_EXCEPTION_H
#define NETWORKINGLIB_EXCEPTION_H

#include <stdexcept>
#include <boost/system/error_code.hpp>
#include <boost/asio/error.hpp>

namespace networking
{
namespace error
{

using ErrorCode = int;

namespace codes
{
constexpr ErrorCode SUCCESS = ErrorCode{0};
constexpr ErrorCode FAILED_OPERATION = ErrorCode{1};
constexpr ErrorCode ABORTED = ErrorCode{2};
constexpr ErrorCode BUSY = ErrorCode{3};
constexpr ErrorCode UNSPECIFIED = ErrorCode{4};
}

class Error : std::runtime_error
{
public:
    Error(const std::string & msg, ErrorCode code)
        : runtime_error(msg), code(code)
    {}

    ErrorCode getErrorCode() const noexcept
    { return code; }

protected:
    ErrorCode code;
};

class FailedOperation : Error
{
public:
    FailedOperation()
        : Error("networking error: failed operation", codes::FAILED_OPERATION)
    {}
};

class Aborted : Error
{
public:
    Aborted()
        : Error("networking error: aborted", codes::ABORTED)
    {}
};

class Busy : Error
{
public:
    Busy()
        : Error("networking error: busy", codes::BUSY)
    {}
};

class Unspecified : Error
{
public:
    Unspecified()
        : Error("networking error: unspecified", codes::UNSPECIFIED)
    {}
};

}
}

#endif //NETWORKINGLIB_EXCEPTION_H
