//
// Created by philipp on 12.01.18.
//

#ifndef NETWORKINGLIB_RESOLVER_H
#define NETWORKINGLIB_RESOLVER_H

#include <memory>
#include <boost/asio/ip/tcp.hpp>
#include "Networking.h"
#include "Error.h"
#include "Utils.h"

namespace networking
{

namespace internal
{

// This class is used for internal purposes. Please use the 'Resolver' class.
// The boost::asio resolvers do not feature a close mechanism which we need in order to perform operations with timeouts.
template<typename Protocol>
class CloseableResolver
{
public:
    using Resolver = typename Protocol::resolver;
    using Query = typename Resolver::query;
    using Iterator = typename Resolver::iterator;

    using ResolveHandler = std::function<void(const boost::system::error_code & error)>;

    CloseableResolver(boost::asio::io_service & ioService)
        : resolver(ioService)
    {}

    void async_resolve(const Query & query,
                       Iterator & endpointIterator,
                       ResolveHandler handler)
    {
        resolver.async_resolve(
            query,
            [&endpointIterator, handler](const auto & error, auto iterator)
            {
                endpointIterator = iterator;
                handler(error);
            });
    }

    void open()
    {
        opened = true;
    }

    bool is_open() const noexcept
    {
        return opened;
    }

    void close(boost::system::error_code &)
    {
        if (!opened)
            return;

        opened = false;
        resolver.cancel();
    }

private:
    std::atomic<bool> opened{true};
    Resolver resolver;
};

}

}

#endif //NETWORKINGLIB_RESOLVER_H
