//
// Created by philipp on 02.12.17.
//

#ifndef PROTOCOL_NETWORKUTILS_H
#define PROTOCOL_NETWORKUTILS_H

#include <boost/asio/ip/tcp.hpp>
#include <boost/asio.hpp>
#include "Networking.h"
#include "Timer.h"
#include "Error.h"
#include "Closeable.h"

namespace networking
{
namespace stream
{

template<typename SyncWriteStream, typename ConstBufferSequence>
void write(Networking & net,
           SyncWriteStream & stream,
           const ConstBufferSequence & buffer,
           const time::Duration & timeout)
{
    auto asyncOperation = [](auto && ... args)
    { boost::asio::async_write(std::forward<decltype(args)>(args)...); };
    closeable::timedOperation(net, asyncOperation, stream, timeout, stream, buffer);
};

template<typename SyncReadStream, typename MutableBufferSequence>
void read(Networking & net,
          SyncReadStream & stream,
          const MutableBufferSequence & buffer,
          const time::Duration & timeout)
{
    auto asyncOperation = [](auto && ... args)
    { boost::asio::async_read(std::forward<decltype(args)>(args)...); };
    closeable::timedOperation(net, asyncOperation, stream, timeout, stream, buffer);
};

template<typename SyncWriteStream, typename MutableBufferSequence, typename WriteHandler>
void asyncWrite(Networking & net,
                SyncWriteStream & stream,
                MutableBufferSequence buffer,
                const time::Duration & timeout,
                WriteHandler handler)
{
    auto asyncOperation = [](auto && ... args)
    { boost::asio::async_write(std::forward<decltype(args)>(args)...); };
    closeable::timedAsyncOperation(net, asyncOperation, stream, timeout, handler, stream, buffer);
}

template<typename SyncReadStream, typename MutableBufferSequence, typename ReadHandler>
void asyncRead(Networking & net,
               SyncReadStream & stream,
               const MutableBufferSequence & buffer,
               const time::Duration & timeout,
               ReadHandler handler)
{
    auto asyncOperation = [](auto && ... args)
    { boost::asio::async_read(std::forward<decltype(args)>(args)...); };
    closeable::timedAsyncOperation(net, asyncOperation, stream, timeout, handler, stream, buffer);
};

}
}

#endif //PROTOCOL_NETWORKUTILS_H
