//
// Created by philipp on 30.11.17.
//

#ifndef PROTOCOL_FOLLOWERVEHICLE_H
#define PROTOCOL_FOLLOWERVEHICLE_H


#include <mutex>
#include <NetworkingLib/DatagramReceiver.h>
#include <NetworkingLib/DatagramSender.h>
#include "Vehicle.h"
#include "VehicleEndpoint.h"
#include "PlatoonMessage.h"
#include "Utils.h"

namespace protocol
{

class FollowerVehicle : public Vehicle
{
private:
    struct PrivateTag
    {
    };

public:
    using Ptr = std::shared_ptr<FollowerVehicle>;

    static Ptr create(networking::Networking & net, VehicleId vehicleId);

    // This should not be used outside.
    FollowerVehicle(PrivateTag, networking::Networking & net, VehicleId vehicleId);

    void stop() override;

    PlatoonConfig getPlatoonConfig() const noexcept;

    void setOnPlatoonConfigUpdatedCallback(Callback callback)
    { onPlatoonConfigUpdatedCallback = callback; }

protected:
    void doCreatePlatoon() override;

    void doRunPlatoon() override;

    void doLeavePlatoon() override;

    void makePlatoonCreateRequest(const VehicleEndpoint & vehicleEndpoint) override;

    PlatoonMessage makePlatoonCreateResponse(
        const tcp::endpoint & clientEndpoint, const PlatoonMessage & request) override;

    Ptr shared_from_this();

private:
    static constexpr networking::time::Duration HEARTBEAT_INTERVAL = std::chrono::milliseconds(50);
    static constexpr networking::time::Duration BROADCAST_TIMEOUT = std::chrono::milliseconds(500);

    VehicleEndpoint leader;
    networking::message::DatagramReceiver<PlatoonMessage>::Ptr broadcastReceiver;
    // We need to track of the timestamp when the last broadcast message arrived.
    // One could think that the timeout-functionality of the networking::message::DatagramReceiver
    // would suffice but we also have to consider non-heartbeat-messages or invalid messages which
    // we could potentially receive as well.
    networking::time::TimePoint lastBroadcastTimestamp;
    networking::message::DatagramSender<PlatoonMessage>::Ptr heartbeatSender;
    networking::time::Timer::Ptr heartbeatTimer;
    Callback onPlatoonConfigUpdatedCallback;

    void startHeartbeat();

    void stopHeartbeat();

    void startReceiveBroadcast();

    void receiveBroadcast();

    void stopReceiveBroadcast();

    void sendLeavePlatoonMessage();
};

}


#endif //PROTOCOL_FOLLOWERVEHICLE_H
