//
// Created by philipp on 01.12.17.
//

#ifndef PROTOCOL_PLATOONCREATEMESSAGE_H
#define PROTOCOL_PLATOONCREATEMESSAGE_H

#include <cstdint>
#include <vector>
#include "Protocol.h"

namespace protocol
{

namespace messageTypes
{
constexpr MessageType LEADER_REQUEST = 0x01;
constexpr MessageType FOLLOWER_REQUEST = 0x02;
constexpr MessageType ACCEPT_RESPONSE = 0x03;
constexpr MessageType REJECT_RESPONSE = 0x04;
constexpr MessageType HEARTBEAT = 0x05;
constexpr MessageType BROADCAST = 0x06;
constexpr MessageType LEAVE_PLATOON = 0x07;
}

union FloatConverter
{
    float floatVal;
    std::uint8_t bytes[4];
};

class PlatoonMessage
{
public:
    using Buffer = std::vector<std::uint8_t>;

    PlatoonMessage(
        VehicleId vehicleId,
        MessageType messageType,
        PlatoonId platoonId,
        PlatoonSpeed platoonSpeed,
        InnerPlatoonDistance innerPlatoonDistance)
        : vehicleId(vehicleId)
          , messageType(messageType)
          , platoonId(platoonId)
          , platoonSpeed(platoonSpeed)
          , innerPlatoonDistance(innerPlatoonDistance)
    {}

    PlatoonMessage(Buffer buffer)
        : vehicleId(0)
          , messageType(0)
          , platoonId(0)
          , platoonSpeed(0)
          , innerPlatoonDistance(0)
    {
        vehicleId += ((VehicleId) buffer[0]);
        vehicleId += ((VehicleId) buffer[1]) << 8;
        vehicleId += ((VehicleId) buffer[2]) << 16;
        vehicleId += ((VehicleId) buffer[3]) << 24;
        messageType += (MessageType) buffer[4];
        platoonId += ((PlatoonId) buffer[5]);
        platoonId += ((PlatoonId) buffer[6]) << 8;
        platoonId += ((PlatoonId) buffer[7]) << 16;
        platoonId += ((PlatoonId) buffer[8]) << 24;
        FloatConverter converter;
        converter.bytes[0] = buffer[9];
        converter.bytes[1] = buffer[10];
        converter.bytes[2] = buffer[11];
        converter.bytes[3] = buffer[12];
        platoonSpeed = converter.floatVal;
        converter.bytes[0] = buffer[13];
        converter.bytes[1] = buffer[14];
        converter.bytes[2] = buffer[15];
        converter.bytes[3] = buffer[16];
        innerPlatoonDistance = converter.floatVal;
    }

    static PlatoonMessage followerRequest(VehicleId vehicleId)
    { return PlatoonMessage(vehicleId, messageTypes::FOLLOWER_REQUEST, 0, 0, 0); }

    static PlatoonMessage leaderRequest(VehicleId vehicleId, PlatoonId platoonId)
    { return PlatoonMessage(vehicleId, messageTypes::LEADER_REQUEST, platoonId, 0, 0); }

    static PlatoonMessage rejectResponse(VehicleId vehicleId)
    { return PlatoonMessage(vehicleId, messageTypes::REJECT_RESPONSE, 0, 0, 0); }

    static PlatoonMessage acceptResponse(VehicleId vehicleId, PlatoonId platoonId)
    { return PlatoonMessage(vehicleId, messageTypes::ACCEPT_RESPONSE, platoonId, 0, 0); }

    static PlatoonMessage broadcastMessage(VehicleId vehicleId,
                                           PlatoonId platoonId,
                                           PlatoonSpeed platoonSpeed,
                                           InnerPlatoonDistance innerPlatoonDistance)
    { return PlatoonMessage(vehicleId, messageTypes::BROADCAST, platoonId, platoonSpeed, innerPlatoonDistance); }

    static PlatoonMessage heartbeatMessage(VehicleId vehicleId, PlatoonId platoonId)
    { return PlatoonMessage(vehicleId, messageTypes::HEARTBEAT, platoonId, 0, 0); }

    static PlatoonMessage leaveMessage(VehicleId vehicleId, PlatoonId platoonId)
    { return PlatoonMessage(vehicleId, messageTypes::LEAVE_PLATOON, platoonId, 0, 0); }

    static Buffer emptyBuffer()
    {
        return Buffer(
            sizeof(vehicleId)
            + sizeof(messageType)
            + sizeof(platoonId)
            + sizeof(platoonSpeed)
            + sizeof(innerPlatoonDistance));
    }

    MessageType getMessageType() const noexcept
    { return messageType; }

    VehicleId getVehicleId() const noexcept
    { return vehicleId; }

    PlatoonId getPlatoonId() const noexcept
    { return platoonId; }

    PlatoonSpeed getPlatoonSpeed() const noexcept
    { return platoonSpeed; }

    InnerPlatoonDistance getInnerPlatoonDistance() const noexcept
    { return innerPlatoonDistance; }

    Buffer data() const
    {
        Buffer data = emptyBuffer();
        data[0] = (std::uint8_t) (vehicleId & 0x000000ff);
        data[1] = (std::uint8_t) ((vehicleId & 0x0000ff00) >> 8);
        data[2] = (std::uint8_t) ((vehicleId & 0x00ff0000) >> 16);
        data[3] = (std::uint8_t) ((vehicleId & 0xff000000) >> 24);
        data[4] = (std::uint8_t) (messageType);
        data[5] = (std::uint8_t) (platoonId & 0x000000ff);
        data[6] = (std::uint8_t) ((platoonId & 0x0000ff00) >> 8);
        data[7] = (std::uint8_t) ((platoonId & 0x00ff0000) >> 16);
        data[8] = (std::uint8_t) ((platoonId & 0xff000000) >> 24);
        FloatConverter converter;
        converter.floatVal = platoonSpeed;
        data[9] = converter.bytes[0];
        data[10] = converter.bytes[1];
        data[11] = converter.bytes[2];
        data[12] = converter.bytes[3];
        converter.floatVal = innerPlatoonDistance;
        data[13] = converter.bytes[0];
        data[14] = converter.bytes[1];
        data[15] = converter.bytes[2];
        data[16] = converter.bytes[3];
        return data;
    }

private:
    VehicleId vehicleId;
    MessageType messageType;
    PlatoonId platoonId;
    PlatoonSpeed platoonSpeed;
    InnerPlatoonDistance innerPlatoonDistance;
};
}

#endif //PROTOCOL_PLATOONCREATEMESSAGE_H
